/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static char rcsid[] =
"@(#) $Header: /n/olympus/opt/users/CVS/ns-1.1-master/tcp-reno.cc,v 1.2 1997/09/08 20:51:16 wuchang Exp $ (LBL)";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "tcp.h"

class RenoTcpAgent : public TcpAgent {
 public:
	RenoTcpAgent();
	virtual int window();
	virtual void recv(Packet *pkt);
	virtual void timeout(int tno);
 protected:
	u_int dupwnd_;
};

static class RenoTcpMatcher : public Matcher {
public:
	RenoTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-reno") == 0)
			return (new RenoTcpAgent());
		return (0);
	}
} matcher_reno;

int RenoTcpAgent::window()
{
	//
	// reno: inflate the window by dupwnd_
	//	dupwnd_ will be non-zero during fast recovery,
	//	at which time it contains the number of dup acks
	//
	int win = int(cwnd_) + dupwnd_;
	if (win > int(wnd_))
		win = int(wnd_);
	return (win);
}

RenoTcpAgent::RenoTcpAgent() : dupwnd_(0)
{
}

void RenoTcpAgent::recv(Packet *pkt)
{
	if (pkt->type_ != PT_ACK) {
		fprintf(stderr,
			"ns: confiuration error: tcp received non-ack\n");
		exit(1);
	}

	if (pkt->flags_ & PF_ECN)
		quench(1);
	if (pkt->seqno_ > last_ack_) {
		dupwnd_ = 0;
		newack(pkt);
                opencwnd();
   	} else if (pkt->seqno_ == last_ack_)  {
		if (++dupacks_ == NUMDUPACKS) {
			/*
			 * Assume we dropped just one packet.
			 * Retransmit last ack + 1
			 * and try to resume the sequence.
			 */
                       /* The line below, for "bug_fix_" true, avoids
                        * problems with multiple fast retransmits after
			* a retransmit timeout.
                        */
			if ( !bug_fix_ || (highest_ack_ > recover_) ||
			    ( recover_cause_ != 2)) {
				recover_cause_ = 1;
				recover_ = maxseq_;
				closecwnd(1);
				reset_rtx_timer(1);
				output(last_ack_ + 1, PF_DUPACK);
                        }
			dupwnd_ = NUMDUPACKS;
		} else if (dupacks_ > NUMDUPACKS)
			++dupwnd_;
	}
	Packet::free(pkt);
	if (trace_)
		plot();

	/*
	 * Try to send more data
	 */

	if (dupacks_ == 0 || dupacks_ > NUMDUPACKS - 1)
		send(0, 0, maxburst_);
}

void RenoTcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		dupwnd_ = 0;
		dupacks_ = 0;
		if (bug_fix_) recover_ = maxseq_;
		TcpAgent::timeout(tno);
	}
}

class MaliciousRenoTcpAgent : public RenoTcpAgent {
 public:
	MaliciousRenoTcpAgent();
	virtual void closecwnd(int how);
 protected:
	double wnd_factor_;
};

static class MaliciousRenoTcpMatcher : public Matcher {
public:
	MaliciousRenoTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-reno-malicious") == 0)
			return (new MaliciousRenoTcpAgent());
		return (0);
	}
} matcher_reno_malicious;

MaliciousRenoTcpAgent::MaliciousRenoTcpAgent()
{
	link_real("ns_tcp_reno_malicious", "window-factor", &wnd_factor_, 0);
}

/*
 * close down the congestion window
 */
void MaliciousRenoTcpAgent::closecwnd(int how)
{
        if (wnd_factor_ == 0) {
		switch (how) {
		case 0:
			/* timeouts, Tahoe dup acks */
			ssthresh_ = int( window() / 2 );
			/* Old code: ssthresh_ = int(cwnd_ / 2); */
			cwnd_ = int(wnd_init_);
			break;
		case 1:
			/* Reno dup acks, or after a recent congestion indication. */
			cwnd_ = window()/2;
			ssthresh_ = int(cwnd_);
			break;
		case 2:
			/* Tahoe dup acks               *
			* after a recent congestion indication */
			cwnd_ = wnd_init_;
			break;
		default:
			abort();
		}
	}
	else {
		cwnd_ = cwnd_ * wnd_factor_;
		ssthresh_ = int(cwnd_);
        }

	fcnt_ = 0;
        count_ = 0;
}
