/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static char rcsid[] =
"@(#) $Header: /n/olympus/opt/users/CVS/ns-1.1-master/tcp-sack1.cc,v 1.3 1997/09/08 20:51:16 wuchang Exp $ (LBL)";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "tcp.h"
#include "node.h"
#include "scoreboard.h"
#include "random.h"
#include "trace.h"

#define TRUE    1
#define FALSE   0
#define RECOVER_DUPACK  1
#define RECOVER_TIMEOUT 2
#define RECOVER_QUENCH  3

class Sack1TcpAgent : public TcpAgent {
 public:
	Sack1TcpAgent();
	virtual int window();
	virtual void recv(Packet *pkt);
	virtual void timeout(int tno);
	virtual void plot();
	void send(int force, int reason, int maxburst);
 protected:
	u_char timeout_;        /* boolean: sent pkt from timeout? */
	u_char fastrecov_;      /* boolean: doing fast recovery? */
	int pipe_;              /* estimate of pipe size (fast recovery) */ 
	ScoreBoard scb_;
};

static class Sack1TcpMatcher : public Matcher {
public:
	Sack1TcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-sack1") == 0)
			return (new Sack1TcpAgent());
		return (0);
	}
} matcher_sack;

int Sack1TcpAgent::window()
{
        return(int(cwnd_ < wnd_ ? cwnd_ : wnd_));
}

Sack1TcpAgent::Sack1TcpAgent() : pipe_(-1), fastrecov_(FALSE)
{
}

void Sack1TcpAgent::recv(Packet *pkt)
{
	int xmit_seqno;
	int dontSend = 0;

	if (pkt->type_ != PT_ACK) {
		fprintf(stderr,
			"ns: confiuration error: tcp received non-ack\n");
		exit(1);
	}

	if (pkt->flags_ & PF_ECN)
		quench(1);
        if (!fastrecov_) {
                /* normal... not fast recovery */
                if ((int)pkt->seqno_ > last_ack_) {
                        /*
                         * regular ACK not in fast recovery... normal
                         */
                        newack(pkt);
                        opencwnd();
                        timeout_ = FALSE;
			scb_.ClearScoreBoard();
                } else if (timeout_ == FALSE)  {
                        if (pkt->seqno_ != last_ack_) {
                                fprintf(stderr, "pkt seq %d should be %d\n" ,
                                        pkt->seqno_, last_ack_);
                                abort();
                        }
			scb_.UpdateScoreBoard (last_ack_, pkt);
                        /*
                         * a duplicate ACK
                         */
                        if (++dupacks_ == NUMDUPACKS) {
                                /*
                                 * Assume we dropped just one packet.
                                 * Retransmit last ack + 1
                                 * and try to resume the sequence.
                                 */
                                if ((highest_ack_ > recover_) ||
                                    (recover_cause_ != RECOVER_TIMEOUT)) {
                                        recover_cause_ = RECOVER_DUPACK;
                                        recover_ = maxseq_;
                                        pipe_ = int(cwnd_) - NUMDUPACKS;
                                        closecwnd(1);
                                        reset_rtx_timer(1);
					fastrecov_ = TRUE;
					scb_.MarkRetran (last_ack_+1);
                                        output(last_ack_ + 1, PF_DUPACK);
                                }
                        }
                }
                if (dupacks_ == 0)
                        send(FALSE, 0, maxburst_);
        } else {
                /* we are in fast recovery */
                --pipe_;
                if ((int)pkt->seqno_ >= recover_) {
                        /* ACK indicates fast recovery is over */
			recover_ = 0;
			fastrecov_ = FALSE;
                        newack(pkt);
                        timeout_ = FALSE;
			scb_.ClearScoreBoard();

                        /* New window: W/2 - K or W/2? */
                } else if ((int)pkt->seqno_ > highest_ack_) {
                        /* Not out of fast recovery yet.
                         * Update highest_ack_, but not last_ack_. */
			--pipe_;
                        highest_ack_ = (int)pkt->seqno_;
		 	scb_.UpdateScoreBoard (highest_ack_, pkt);
                        t_backoff_ = 1;
                        newtimer(pkt);
                } else if (timeout_ == FALSE) {
                        /* got another dup ack */
			scb_.UpdateScoreBoard (last_ack_, pkt);
                        if (dupacks_ > 0)
                                dupacks_++;
                }
                send(FALSE, 0, maxburst_);
        }

	Packet::free(pkt);
	if (trace_)
		plot();
}

void Sack1TcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		dupacks_ = 0;
		fastrecov_ = FALSE;
		timeout_ = TRUE;
		if (highest_ack_ > last_ack_)
			last_ack_ = highest_ack_;
#ifdef DEBUGSACK1A
		printf ("timeout. highest_ack: %d seqno: %d\n", 
		  highest_ack_, seqno_);
#endif
		recover_ = maxseq_;
		scb_.ClearScoreBoard();
	}
	TcpAgent::timeout(tno);
}

void Sack1TcpAgent::send(int force, int reason, int maxburst)
{
        register int pktno, found, nextpktno, npacket = 0;
        int win = window();
	int xmit_seqno;

        found = 1;
        /*
         * as long as the pipe is open and there is app data to send...
         */
        while (((!fastrecov_  && (seqno_ <= last_ack_ + win)) ||
            (fastrecov_ && (pipe_ < int(cwnd_)))) 
		&& seqno_ < curseq_ && found) {

                if (overhead_ == 0 || force) {
                        found = 0;

			xmit_seqno = scb_.GetNextRetran ();
#ifdef DEBUGSACK1A
			printf("highest_ack: %d xmit_seqno: %d\n", 
			highest_ack_, xmit_seqno);
#endif
			if (xmit_seqno == -1) { 
			    if ((!fastrecov_ && seqno_<=highest_ack_+win)||
			     (fastrecov_ && seqno_<=last_ack_+int(2*wnd_))){ 
				found = 1;
				xmit_seqno = seqno_++;
#ifdef DEBUGSACK1A
				printf("sending %d fastrecovery: %d win %d\n",
				  xmit_seqno, fastrecov_, win);
#endif
			    }
			}
			else if (recover_>0 &&
				xmit_seqno<=last_ack_+int(2*wnd_)) {
				found = 1;
				scb_.MarkRetran (xmit_seqno);
				win = window();
			}
			if (found) {
                        	output(xmit_seqno, reason);
                                if (seqno_ <= xmit_seqno)
                                        seqno_ = xmit_seqno + 1;
				npacket++;
				pipe_++;
			}
                } else if (!pending_[TCP_TIMER_DELSND]) {
                        /*
                         * Set a delayed send timeout.
			 * This is only for the simulator,to add some
			 *   randomization if speficied.
                         */
                        sched(Random::uniform(overhead_), TCP_TIMER_DELSND);
                        return;
                }
                if (maxburst && npacket == maxburst)
                        break;
        } /* while */
}

void Sack1TcpAgent::plot()
{
	double t = Scheduler::instance().clock();
	sprintf(trace_->buffer(), "t %g %d rtt %g\n", 
		t, class_, t_rtt_ * tcp_tick_);
	trace_->dump();
	sprintf(trace_->buffer(), "t %g %d dev %g\n", 
		t, class_, t_rttvar_ * tcp_tick_);
	trace_->dump();
	sprintf(trace_->buffer(), "t %g %d win %f\n", t, class_, cwnd_);
	trace_->dump();
	sprintf(trace_->buffer(), "t %g %d bck %d\n", t, class_, t_backoff_);
	trace_->dump();
}

// Sack1Police, Sack1Cload, Sack1Timed, Sack1TimedCload by
// Wu-chang Feng <wuchang@eecs.umich.edu>

class Sack1PoliceTcpAgent : public Sack1TcpAgent {
 public:
	Sack1PoliceTcpAgent();
	virtual void reset();
	virtual void output(int seqno, int reason);
	void send(int force, int reason, int maxburst = 0);
	void update_bucket();
 protected:
	double lbrate_;         /* Leaky bucket rate (bits/sec) */
	double lbdepth_;        /* Leaky bucket depth (seconds) */
	double timer_interval_; /* Fine-grained timer-interval */
	int compliant_;
	int non_compliant_;
	double lbdepth_bytes_;	 // Leaky bucket depth (bytes)
	double last_arriv_;      // Last time a packet arrived
	double tokens_avail_;    // Tokens left (bytes)
	double tokens_wasted_;   // Tokens wasted (bytes)
	void plot();
};

static class Sack1PoliceTcpMatcher : public Matcher {
public:
	Sack1PoliceTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-sack1-police") == 0)
			return (new Sack1PoliceTcpAgent());
		return (0);
	}
} matcher_sackpolice;

Sack1PoliceTcpAgent::Sack1PoliceTcpAgent()
{
	link_bw("ns_tcp_sack1_police", "lbrate", &lbrate_, 0);
	link_time("ns_tcp_sack1_police","lbdepth",&lbdepth_, 0);
	link_time("ns_tcp_sack1_police","timer-interval",&timer_interval_, 0);

	link_real(0, "lbdepth-bytes", &lbdepth_bytes_, 1);
	link_real(0, "last-arrival", &last_arriv_, 1);
	link_real(0, "tokens", &tokens_avail_, 1);
	link_real(0, "tokens-wasted", &tokens_wasted_, 1);
	link_int(0, "compliant", &compliant_, 1);
	link_int(0, "non-compliant", &non_compliant_,1);
}

void Sack1PoliceTcpAgent::reset()
{
	TcpAgent::reset();

	last_arriv_ = 0;
	tokens_avail_ = 0;
	tokens_wasted_ = 0;
	compliant_ = 0;
	non_compliant_ = 0;
	lbdepth_bytes_ = lbrate_ * lbdepth_ / 8;
}

void Sack1PoliceTcpAgent::update_bucket()
{
	double now = Scheduler::instance().clock();

	// This is to allow you to change the lbdepth on the fly
	// Yes, I know this is inefficient.

	lbdepth_bytes_ = (lbrate_/8) * lbdepth_;

	tokens_avail_ += double((now-last_arriv_)*lbrate_)/8.0;
	last_arriv_ = now;
	if (tokens_avail_ > lbdepth_bytes_ ) {
		tokens_wasted_ += (tokens_avail_ - lbdepth_bytes_ );
		tokens_avail_ = lbdepth_bytes_ ;
		if (trace_) {
			sprintf(trace_->buffer(),"loss %g %5.0f",now,tokens_wasted_ );
			trace_->dump();
		}
	}
}

/*
 * Called from DataSource
 */
void Sack1PoliceTcpAgent::output(int seqno, int reason)
{
        Packet* p = allocpkt(seqno);
        double now = Scheduler::instance().clock();

        update_bucket();

        // check to see if packet is compliant
        // mark packet if compliant
        if (p->size_ <= tokens_avail_) {
                p->flags_ |= PF_PRI;
                tokens_avail_ = tokens_avail_ - p->size_;
                compliant_++;
        } else {
                non_compliant_++;
        }

        p->bd_.tcp_.ts_ = now;
        p->flags_ |= reason;
        node_->transmit(p);
        if (seqno > maxseq_) {
                maxseq_ = seqno;
                if (!rtt_active_) {
                        rtt_active_ = 1;
                        if (seqno > rtt_seq_)
                                rtt_seq_ = seqno;
                }
        }
        if (!pending_[TCP_TIMER_RTX])
                /* No timer pending.  Schedule one. */
                set_rtx_timer();
}


void Sack1PoliceTcpAgent::send(int force, int reason, int maxburst)
{
        register int pktno, found, nextpktno, npacket = 0;
        int win = window();
        int xmit_seqno;

	update_bucket();
        found = 1;
        /*
         * as long as the pipe is open and there is app data to send...
         */
        while (((!fastrecov_  && (seqno_ <= last_ack_ + win)) ||
            (fastrecov_ && (pipe_ < int(cwnd_))))
                && seqno_ < curseq_ && found) {

                if (overhead_ == 0 || force || (tokens_avail_ > size_) || pending_[TCP_TIMER_DELSND]) {
                        found = 0;

                        xmit_seqno = scb_.GetNextRetran ();
#ifdef DEBUGSACK1A
                        printf("highest_ack: %d xmit_seqno: %d\n",
                        highest_ack_, xmit_seqno);
#endif
                        if (xmit_seqno == -1) {
                            if ((!fastrecov_ && seqno_<=highest_ack_+win)||
                             (fastrecov_ && seqno_<=last_ack_+int(2*wnd_))){
                                found = 1;
                                xmit_seqno = seqno_++;
#ifdef DEBUGSACK1A
                                printf("sending %d fastrecovery: %d win %d\n",
                                  xmit_seqno, fastrecov_, win);
#endif
                            }
                        }
                        else if (recover_>0 &&
                                xmit_seqno<=last_ack_+int(2*wnd_)) {
                                found = 1;
                                scb_.MarkRetran (xmit_seqno);
                                win = window();
                        }
                        if (found) {
                                output(xmit_seqno, reason);
                                if (seqno_ <= xmit_seqno)
                                        seqno_ = xmit_seqno + 1;
                                npacket++;
                                pipe_++;
                        }
                } else if (!pending_[TCP_TIMER_DELSND]) {
                        /*
                         * Set a delayed send timeout.
                         * This is only for the simulator,to add some
                         *   randomization if speficied.
                         */
                        sched(Random::uniform(overhead_), TCP_TIMER_DELSND);
                        return;
                }
                if (maxburst && npacket == maxburst)
                        break;
        } /* while */
}

void Sack1PoliceTcpAgent::plot() 
{
	double t = Scheduler::instance().clock();
	sprintf(trace_->buffer(), "%g %f %d %f %d", t, cwnd_, last_ack_, cwnd_ - seqno_ + last_ack_ , seqno_ );
	trace_->dump();
}

class Sack1CloadTcpAgent : virtual public Sack1PoliceTcpAgent {
 public:
	Sack1CloadTcpAgent();
	virtual void closecwnd(int how);
	virtual void opencwnd();
 protected:
};

static class Sack1CloadTcpMatcher : public Matcher {
public:
	Sack1CloadTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-sack1-cload") == 0)
			return (new Sack1CloadTcpAgent());
		return (0);
	}
} matcher_sackcload;

Sack1CloadTcpAgent::Sack1CloadTcpAgent()
{
}


/*
 * open up the congestion window
 */
void Sack1CloadTcpAgent::opencwnd()
{
	double rbcwnd, prorate, above_rbcwnd;

	switch (wnd_option_) {
	case 0:
		rbcwnd = double( lbrate_ * t_brtt_ * tcp_tick_ / (8*size_) );
		break;
	case 1:
	case 2:
		rbcwnd = double( lbrate_ * t_rtt_ * tcp_tick_ / (8*size_) );
		break;
	default:
		abort();
	}

	if (cwnd_ < (rbcwnd + 1)) {
		cwnd_ = rbcwnd + 1;
		above_rbcwnd = 1;
		return;
	}
	else {
		above_rbcwnd = cwnd_ - rbcwnd;
		prorate = double( above_rbcwnd / cwnd_ );
	}

	if (cwnd_ < ssthresh_) {
		switch (wnd_option_) {
		case 0:
		case 1:
			cwnd_ += prorate;
			break;
		case 2:
			/* slow-start (exponential) */
			cwnd_ += 1;
			break;
		}
	} else {
		/* linear */
		switch (wnd_option_) {
		case 0:
		case 1:
		case 2:
			/* This is the standard algorithm. */
			cwnd_ += 1 / cwnd_;
			break;

		default:
#ifdef notdef
			/*XXX*/
			error("illegal window option %d", wnd_option_);
#endif
			abort();
		}
	}
	// if maxcwnd_ is set (nonzero), make it the cwnd limit
	if (maxcwnd_ && (int(cwnd_) > maxcwnd_))
		cwnd_ = maxcwnd_;

	return;
}

/*
 * close down the congestion window
 */
void Sack1CloadTcpAgent::closecwnd(int how)
{
        double rbcwnd, above_rbcwnd;  /* rate based cwnd */

	switch (wnd_option_) {
	case 0:
		rbcwnd = double( lbrate_ * t_brtt_ * tcp_tick_ / (8*size_) );
		break;
	case 1:
	case 2:
		rbcwnd = double( lbrate_ * t_rtt_ * tcp_tick_ / (8*size_) );
		break;
	default:
		abort();
	}

	if ( cwnd_ < rbcwnd ) {
		cwnd_ = rbcwnd + 1;
		return;
	}
	else {
		above_rbcwnd = cwnd_ - rbcwnd ;
		switch (how) {
		case 0:
			ssthresh_ = int(rbcwnd + above_rbcwnd/2);
			cwnd_ = rbcwnd + 1;
			break;
			/* timeout: set cwnd to rate based window.
			*          set ssthresh to rbwnd+(cwnd-rbwnd)/2
			*/
		case 1:
			cwnd_ = rbcwnd + above_rbcwnd/2;
			ssthresh_ = int(cwnd_);
			break;
		case 2:
			/* Tahoe dup acks: set cwnd to rate based window.
			* No congestion avoidance (ssthresh) */
			cwnd_ = rbcwnd + 1;
			break;
		default:
			abort();
		}
	}
	fcnt_ = 0.;
	count_ = 0;
}

class Sack1TimedTcpAgent : virtual public Sack1PoliceTcpAgent {
 public:
	Sack1TimedTcpAgent();
	virtual void reset();
	virtual void timeout(int tno);
	void send(int force, int reason, int maxburst = 0);
 protected:
	void set_rbsnd_timer();
};

static class Sack1TimedTcpMatcher : public Matcher {
public:
	Sack1TimedTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-sack1-timed") == 0)
			return (new Sack1TimedTcpAgent());
		return (0);
	}
} matcher_sacktimed;

Sack1TimedTcpAgent::Sack1TimedTcpAgent()
{
}

void Sack1TimedTcpAgent::reset()
{
	Sack1PoliceTcpAgent::reset();
	set_rbsnd_timer();
}

void Sack1TimedTcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		dupacks_ = 0;
		fastrecov_ = FALSE;
		timeout_ = TRUE;
		if (highest_ack_ > last_ack_)
			last_ack_ = highest_ack_;
#ifdef DEBUGSACK1A
		printf ("timeout. highest_ack: %d seqno: %d\n", 
		  highest_ack_, seqno_);
#endif
		recover_ = maxseq_;
		scb_.ClearScoreBoard();
	}
        	/* retransmit timer */
	if (tno == TCP_TIMER_RTX) {
		recover_ = maxseq_;
		recover_cause_ = 2;
		closecwnd(0);
		reset_rtx_timer(0);
		send(0, PF_TIMEOUT);
	} else if (tno == TCP_TIMER_DELSND) {
		/*
		 * delayed-send timer, with random overhead
		 * to avoid phase effects
		 */
		send(1, PF_TIMEOUT);
	} else if (tno == TCP_TIMER_RBSND) {
		update_bucket();
		if (trace_) {
			sprintf(trace_->buffer(), "t %g tok=%g", 
				Scheduler::instance().clock(), tokens_avail_);
			trace_->dump();
		}
                send(1, PF_RBSND);
                if (pending_[TCP_TIMER_RBSND]) {
                  fprintf(stderr,"rbsnd timer should have been cancelled\n");
                  abort();
                }
                set_rbsnd_timer();
        }

}

void Sack1TimedTcpAgent::set_rbsnd_timer()
{
	if (pending_[TCP_TIMER_RBSND]) {
		cancel(TCP_TIMER_RBSND);
	}
	sched(timer_interval_,TCP_TIMER_RBSND);
}

void Sack1TimedTcpAgent::send(int force, int reason, int maxburst)
{
        register int pktno, found, nextpktno, npacket = 0;
        int win = window();
	int xmit_seqno;

        update_bucket();
        found = 1;
        /*
         * as long as the pipe is open and there is app data to send...
         */
        while (
		((reason == PF_RBSND) && (tokens_avail_ > size_) && (seqno_ <= last_ack_ + wnd_) && (seqno_ < curseq_) && found) ||
		((reason != PF_RBSND) && fastrecov_ && (pipe_ < int(cwnd_)) && (seqno_ < curseq_) && found) ||
		((reason != PF_RBSND) && !fastrecov_ && (seqno_ <= last_ack_ + win) && (seqno_ < curseq_) && found)
		)
	{
                if (overhead_ == 0 || force || (tokens_avail_ > size_) || pending_[TCP_TIMER_DELSND]) {
                        found = 0;

			xmit_seqno = scb_.GetNextRetran ();
#ifdef DEBUGSACK1A
			printf("highest_ack: %d xmit_seqno: %d\n", 
			highest_ack_, xmit_seqno);
#endif
			if (xmit_seqno == -1) { 
			    if ((!fastrecov_ && seqno_<=highest_ack_+win)||
			     (fastrecov_ && seqno_<=last_ack_+int(2*wnd_)) ||
                             ((reason == PF_RBSND) && (seqno_ <= last_ack_+wnd_)))
                            { 
				found = 1;
				xmit_seqno = seqno_++;
#ifdef DEBUGSACK1A
				printf("sending %d fastrecovery: %d win %d\n",
				  xmit_seqno, fastrecov_, win);
#endif
			    }
			}
			else if (recover_>0 &&
				xmit_seqno<=last_ack_+int(2*wnd_)) {
				found = 1;
				scb_.MarkRetran (xmit_seqno);
				win = window();
			}
			if (found) {
                        	output(xmit_seqno, reason);
                                if (seqno_ <= xmit_seqno)
                                        seqno_ = xmit_seqno + 1;
				npacket++;
				pipe_++;
			}
                } else if (!pending_[TCP_TIMER_DELSND]) {
                        /*
                         * Set a delayed send timeout.
			 * This is only for the simulator,to add some
			 *   randomization if speficied.
                         */
                        sched(Random::uniform(overhead_), TCP_TIMER_DELSND);
                        return;
                }
                if (maxburst && npacket == maxburst)
                        break;
        } /* while */
}

class Sack1TimedCloadTcpAgent : public Sack1TimedTcpAgent, public Sack1CloadTcpAgent {
 public:
	Sack1TimedCloadTcpAgent();
 protected:
};

static class Sack1TimedCloadTcpMatcher : public Matcher {
public:
	Sack1TimedCloadTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-sack1-timed-cload") == 0)
			return (new Sack1TimedCloadTcpAgent());
		return (0);
	}
} matcher_sacktimedcload;

Sack1TimedCloadTcpAgent::Sack1TimedCloadTcpAgent()
{
}

class MaliciousSack1TcpAgent : public Sack1TcpAgent {
 public:
        MaliciousSack1TcpAgent();
        virtual void closecwnd(int how);
 protected:
        double wnd_factor_;
};

static class MaliciousSack1TcpMatcher : public Matcher {
public:
        MaliciousSack1TcpMatcher() : Matcher("agent") {}
        TclObject* match(const char* type) {
                if (strcmp(type, "tcp-sack1-malicious") == 0)
                        return (new MaliciousSack1TcpAgent());
                return (0);
        }
} matcher_sack1_malicious;

MaliciousSack1TcpAgent::MaliciousSack1TcpAgent()
{
        link_real("ns_tcp_sack1_malicious", "window-factor", &wnd_factor_, 0);
}

/*
 * close down the congestion window
 */
void MaliciousSack1TcpAgent::closecwnd(int how)
{
        if (wnd_factor_ == 0) {
                switch (how) {
                case 0:
                        /* timeouts, Tahoe dup acks */
                        ssthresh_ = int( window() / 2 );
                        /* Old code: ssthresh_ = int(cwnd_ / 2); */
                        cwnd_ = int(wnd_init_);
                        break;
                case 1:
                        /* Reno dup acks, or after a recent congestion indicatio
n. */
                        cwnd_ = window()/2;
                        ssthresh_ = int(cwnd_);
                        break;
                case 2:
                        /* Tahoe dup acks               *
                        * after a recent congestion indication */
                        cwnd_ = wnd_init_;
                        break;
                default:
                        abort();
                }
        }
        else {
                cwnd_ = cwnd_ * wnd_factor_;
		ssthresh_ = int(cwnd_);
        }

        fcnt_ = 0;
        count_ = 0;
}
